"""Exit task for ScriptEngine."""

from scriptengine.tasks.core import Task
from scriptengine.exceptions import ScriptEngineStopException


class Exit(Task):
    """Exit task, run method throws ScriptEngineStopException
    """
    def run(self, context):
        self.log_info(
            self.getarg(
                'msg',
                context,
                default='Requesting ScriptEngine to STOP'
            )
        )
        raise ScriptEngineStopException



import sys
from copy import deepcopy

from deepdiff import DeepDiff, Delta
from deepmerge import always_merger

_python_version = (sys.version_info.major, sys.version_info.minor)


def save_copy(context):
    """For versions other than 3.6, this is just copy.deepcopy.
    For 3.6, we have to remove ["se"]["instance"] before deepcopying and restore
    afterwards, because this constitutes some recursive reference that 3.6 has a
    problem with.
    """
    if _python_version != (3, 6):
        return deepcopy(context)
    singleton = object()
    try:
        se_instance = context["se"]["instance"]
        context["se"]["instance"] = None
    except KeyError:
        se_instance = singleton
    copied_context = deepcopy(context)
    if se_instance is not singleton:
        copied_context["se"]["instance"] = context["se"]["instance"] = se_instance
    return copied_context


def context_delta(first, second):
    """The deepdiff.Delta's __init__ and __add__ functions use copy.deepcopy
    internally, which does not work with the SE context (see above).
    Therefore, we ignore context["se"]["instance"] before creating Delta. This
    means, that for Python 3.6 this item in the context is not part of the Delta,
    even if it was changed between first and second!
    Also, we have to create the Delta with mutate=True to avoid another deepcopy.
    """
    if _python_version != (3, 6):
        return Delta(DeepDiff(first, second))

    singleton = object()
    try:
        first_se_instance = first["se"]["instance"]
        first["se"]["instance"] = None
    except KeyError:
        first_se_instance = singleton
    try:
        second_se_instance = second["se"]["instance"]
        second["se"]["instance"] = None
    except KeyError:
        second_se_instance = singleton
    delta = Delta(DeepDiff(first, second), mutate=True)
    if first_se_instance is not singleton:
        first["se"]["instance"] = first_se_instance
    if second_se_instance is not singleton:
        second["se"]["instance"] = second_se_instance
    return delta


class Context(dict):
    pass


class ContextUpdate:
    def __init__(self, base=None, second=None):
        if base is None:
            self.merge = self.delta = None
        elif second is None:
            if not isinstance(base, dict):
                raise TypeError(
                    "First argument of ContextUpdate() must be None or a dict"
                )
            self.merge = base
            self.delta = None
        else:
            self.merge = None
            self.delta = context_delta(base, second)

    def __radd__(self, other):
        if isinstance(other, dict):
            if self.merge:
                return always_merger.merge(other, self.merge)
            elif self.delta:
                return other + self.delta
            return other
        raise TypeError(
            f"Unsupported operand types for +: '{type(other)}' and 'ContextUpdate'"
        )

    def __repr__(self):
        if self.merge:
            return f"<ContextUpdate: merge {self.merge}>"
        elif self.delta:
            return f"<ContextUpdate: {self.delta}>"
        else:
            return "<ContextUpdate: None>"



"""Include task for ScriptEngine

   The Include task reads a YAML file, parses the content to create a
   ScriptEngine script and lets the active ScriptEngine instance execute it.
"""

from pathlib import Path

from scriptengine.context import Context, ContextUpdate, save_copy
from scriptengine.exceptions import ScriptEngineTaskRunError
from scriptengine.tasks.core import Task, timed_runner
from scriptengine.yaml.parser import parse_file


def _file_in_paths(file, paths=None):
    if file.is_absolute():
        if file.is_file():
            return file
        raise FileNotFoundError

    searched_paths = []
    for p in paths or []:
        if p not in searched_paths:
            searched_paths.append(p)
            fpath = p / file
            if fpath.is_file():
                return fpath
    raise FileNotFoundError


class Include(Task):

    _required_arguments = ("src",)

    def __init__(self, arguments):
        Include.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f"Include: {self.src}"

    @timed_runner
    def run(self, context):

        src = Path(self.getarg("src", context))
        self.log_info(f"Include script from {src}")

        # Search for include file in '.', old cwd, and script_path
        cwd = Path(".")
        ocwd = Path(context["se"]["cli"]["cwd"])
        script_path = (Path(p) for p in context["se"]["cli"]["script_path"])
        search_path = (cwd, ocwd, *script_path)
        self.log_debug(
            f"Include file search path: {tuple(str(p) for p in search_path)}"
        )

        try:
            inc_file = _file_in_paths(src, search_path)
        except FileNotFoundError:
            if self.getarg("ignore_not_found", default=False):
                self.log_warning(f"Include file not found: {src}")
                return
            self.log_error(f"Include file not found: {src}")
            raise ScriptEngineTaskRunError
        self.log_debug(f"Include file found: {str(inc_file)}")

        script = parse_file(inc_file)

        self.log_debug(f"Execute include script: {inc_file}")
        local_context = Context(save_copy(context))
        context_update = local_context["se"]["instance"].run(script, local_context)
        if context_update:
            local_context += context_update
        self.log_debug(f"Finished executing include script: {inc_file}")

        return ContextUpdate(context, local_context)



"""ScriptEngine exceptions
"""


class ScriptEngineStopException(Exception):
    """Raise this exception when ScriptEngine should stop"""


class ScriptEngineError(Exception):
    """All ScriptEngine errors should throw this"""


class ScriptEngineParseError(ScriptEngineError):
    """An error that has to do with parsing SE scripts"""


class ScriptEngineParseFileError(ScriptEngineParseError):
    """An file error while acessing an SE script"""


class ScriptEngineParseJinjaError(ScriptEngineParseError):
    """Throw this exception whenever there is a problem with parsing Jinja2
    expressions"""


class ScriptEngineParseYAMLError(ScriptEngineParseError):
    """Throw this exception whenever there is a problem with parsing YAML"""


class ScriptEngineParseScriptError(ScriptEngineParseError):
    """Throw this exception if YAML/Jinja2 is okay, but the result can not be
    parsed as a valid ScriptEngine script"""


class ScriptEngineTaskError(ScriptEngineError):
    """All ScriptEngine Tasks should throw this"""


class ScriptEngineTaskLoaderError(ScriptEngineTaskError):
    """There is a problem while loading an SE task or taskset"""


class ScriptEngineTaskArgumentError(ScriptEngineTaskError):
    """There is a problem with an SE Task argument"""


class ScriptEngineTaskArgumentInvalidError(ScriptEngineTaskArgumentError):
    """An SE Task argument is not valid"""


class ScriptEngineTaskArgumentMissingError(ScriptEngineTaskArgumentError):
    """An SE Task argument is missing"""


class ScriptEngineTaskRunError(ScriptEngineTaskError):
    """Throw this for errors at run time, i.e. during execution of the tasks
    run() method"""


class ScriptEngineJobError(ScriptEngineError):
    """All ScriptEngine Jobs should throw this"""


class ScriptEngineJobParseError(ScriptEngineJobError):
    """An error while parsing a job description"""



"""ScriptEngine YAML parsing
"""

import logging

import dateutil.rrule
import yaml

from scriptengine.exceptions import (
    ScriptEngineParseFileError,
    ScriptEngineParseScriptError,
    ScriptEngineParseYAMLError,
)
from scriptengine.jobs import Job
from scriptengine.tasks.core.loader import load_and_register
from scriptengine.yaml.noparse_strings import (
    NoParseJinjaString,
    NoParseString,
    NoParseYamlString,
)


def string_class_constructor(derived_string_class):
    """YAML constructor factory. The constructors convert their values to a
    string and return an object of a specific class (derived from str). The
    class membership is later used to limit the Jinja2/YAML parsing during
    the processing of ScriptEngine scripts.
    """

    def constructor(loader, node):
        value = loader.construct_scalar(node)
        return derived_string_class(str(value))

    return constructor


yaml.add_constructor("!noparse", string_class_constructor(NoParseString))
yaml.add_constructor("!noparse_yaml", string_class_constructor(NoParseYamlString))
yaml.add_constructor("!noparse_jinja", string_class_constructor(NoParseJinjaString))


def rrule_constructor(loader, node):
    """A YAML constructor that can parse RFC5545 rrules"""
    value = loader.construct_scalar(node)
    rrule = dateutil.rrule.rrulestr(value)
    return rrule


yaml.add_constructor("!rrule", rrule_constructor)


def parse(data):
    """Recursively parses data and returns a ScriptEnginge Task or Job, or a list of those.
    The data is supposed to come from YAML-parsing a ScriptEngine script."""

    def build_job(todo, spec):
        sentinel = object()

        when_clause = spec.get("when")
        loop_descriptor = spec.get("loop", sentinel)

        loop_opts = {}
        if loop_descriptor is not sentinel:
            if not loop_descriptor:
                log.error(
                    f"Invalid loop descriptor (evaluates to False): {loop_descriptor}"
                )
                raise ScriptEngineParseYAMLError
            if isinstance(loop_descriptor, list) or isinstance(loop_descriptor, str):
                loop_opts = {
                    "loop": loop_descriptor,
                    "loop_vars": None,
                }
            elif isinstance(loop_descriptor, dict) and "in" in loop_descriptor:
                loop_opts = {
                    "loop": loop_descriptor["in"],
                    "loop_vars": loop_descriptor.get("with"),
                }
            else:
                log.error(f"Invalid loop descriptor: {loop_descriptor}")
                raise ScriptEngineParseYAMLError
        return Job(todo, when=when_clause, **loop_opts)

    if not data:
        return []

    if isinstance(data, list):
        return [parse(d) for d in data]

    log = logging.getLogger("se.yaml")

    tasks = load_and_register()
    job_keys = {"do"}

    try:
        valid_keys = data.keys() & (job_keys | tasks.keys())
    except AttributeError:
        log.error(f"Expected YAML list or dict, got {type(data).__name__}: {data}")
        raise ScriptEngineParseScriptError

    if not valid_keys:
        log.error(f"No valid task name found in {data.keys()}")
        raise ScriptEngineParseScriptError

    if len(valid_keys) > 1:
        log.error(f"Ambiguous task names found in {data.keys()}")
        raise ScriptEngineParseScriptError

    key = valid_keys.pop()

    if key in tasks:
        if len(data) == 1:  # It's a simple Task
            return tasks[key](data[key])
        return build_job(parse({key: data[key]}), data)

    return build_job(
        [parse(t) for t in data[key]],
        data,
    )


def parse_file(filename):
    """Reads a ScriptEngine script from a YAML file.

    Args:
        filename (str): Path to YAML file

    Returns:
        A scriptengine.task.Task, a scriptengine.jobs.Job, or a list of
        tasks/jobs.
    """
    try:
        with open(filename) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
    except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
        logging.getLogger("se.yaml").error(f"Could not read script file: {e}")
        raise ScriptEngineParseFileError
    except yaml.YAMLError as e:
        logging.getLogger("se.yaml").error(f"Could not parse script file: {e}")
        raise ScriptEngineParseYAMLError
    return parse(data)



""" SimpleScriptEngine for ScriptEngine.

SimpleScriptEngine is a simplistic script engine. It runs all jobs sequentially
on the local machine, without consideration of job contexts. The
SimpleScriptEngine relies on the Job class to run the actual tasks/jobs.
"""

import logging
import sys
from pprint import pprint

from scriptengine.context import Context, ContextUpdate, save_copy
from scriptengine.exceptions import (
    ScriptEngineJobError,
    ScriptEngineStopException,
    ScriptEngineTaskError,
)
from scriptengine.tasks.core import Task


class SimpleScriptEngine:
    def __init__(self):
        self.logger = logging.getLogger(
            "se.instance." + self.__class__.__name__.lower()
        )

    def _guarded_run(self, runner, context):
        try:
            return runner.run(context)
        except AttributeError as e:
            self.log_error(
                "STOPPING SimpleScriptEngine due to internal error: "
                f"Cannot run type {type(runner).__name__}"
            )
            error = e
        except ScriptEngineStopException:
            self.log_info("STOPPING SimpleScriptEngine instance upon request")
            error = None
        except ScriptEngineTaskError as e:
            if isinstance(runner, Task):
                self.log_error(
                    "STOPPING SimpleScriptEngine due to task error in "
                    f"{runner.reg_name} <{runner.shortid}>"
                )
            else:
                self.log_error(
                    "STOPPING SimpleScriptEngine due to task error in job "
                    f"<{runner.shortid}>"
                )
            error = e
        except ScriptEngineJobError as e:
            self.log_error(
                f"STOPPING SimpleScriptEngine due to job error in <{runner.shortid}>"
            )
            error = e
        if error:
            if self.logger.getEffectiveLevel() <= logging.DEBUG:
                self.log_error("Last context before error:")
                pprint(context)
                self.log_error("Traceback from error:")
                raise error
            else:
                self.log_error("For more debugging info, re-run with loglevel DEBUG")
        sys.exit()

    def run(self, script, context):
        local_context = Context(save_copy(context))
        for todo in script if isinstance(script, list) else [script]:
            context_update = self._guarded_run(todo, local_context)
            if context_update:
                local_context += context_update
        return ContextUpdate(context, local_context)

    def _log(self, level, msg):
        self.logger.log(level, msg)

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



class NoParseYamlString(str):
    """This represents a string that should not be YAML-parsed"""


class NoParseJinjaString(str):
    """This represents a string that should not be parsed with Jinja2"""


class NoParseString(NoParseYamlString, NoParseJinjaString):
    """This represents a string that should not be parsed at all, neither YAML
    or Jinja2"""



"""Find task for ScriptEngine"""

import fnmatch
import os

from scriptengine.context import ContextUpdate
from scriptengine.exceptions import ScriptEngineTaskArgumentInvalidError
from scriptengine.tasks.core import Task, timed_runner


class Find(Task):
    """Find task, finds files or directories by name patterns"""

    _required_arguments = ("path",)

    def __init__(self, arguments):
        Find.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        path = os.path.normpath(self.getarg("path", context))
        path_depth = path.count(os.path.sep)

        pattern = self.getarg("pattern", context, default="*")

        find_type = self.getarg("type", context, default="file")
        if find_type not in ("file", "dir"):
            msg = (
                f'Invalid "type" argument '
                f'(must be either "file" or "dir"): {find_type}'
            )
            self.log_error(msg)
            raise ScriptEngineTaskArgumentInvalidError(msg)

        max_depth = self.getarg("depth", context, default=-1)
        try:
            max_depth = max(-1, int(max_depth))
        except (ValueError, TypeError):
            msg = f'Invalid "depth" argument (not an integer): {max_depth}'
            self.log_error(msg)
            raise ScriptEngineTaskArgumentInvalidError(msg)

        self.log_info(
            f'Find {find_type} with pattern "{pattern}" '
            f"in {path} (with max depth={max_depth})"
        )
        result = []
        for root, dirs, files in os.walk(path):
            if find_type == "file":
                files_or_dirs = files
            else:
                files_or_dirs = dirs
            for name in files_or_dirs:
                if fnmatch.fnmatch(name, pattern):
                    result.append(os.path.normpath(os.path.join(root, name)))
            if max_depth >= 0 and root.count(os.path.sep) >= path_depth + max_depth:
                del dirs[:]
        if result:
            self.log_debug(f"Found: {result}")
        else:
            self.log_debug("Nothing found")
        result_key = self.getarg("set", context, default="result")
        self.log_debug(f'Store result under context key "{result_key}"')

        return ContextUpdate({result_key: result})



"""Time task for ScriptEngine, provides the current absolute time or elapsed
   time since a reference in the context.
"""

import datetime

from scriptengine.tasks.core import Task, timed_runner


class Time(Task):
    """Task class for measuring time in ScriptEngine
    """
    _required_arguments = ('set', )

    def __init__(self, arguments):
        Time.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):

        key = self.getarg("set", context)
        since = self.getarg("since", context, default=None)

        if since:
            self.log_info(f"Storing time delta in '{key}'")
        else:
            self.log_info(f"Storing datetime in '{key}'")

        now = datetime.datetime.now()

        if since:
            context[key] = now - since
        else:
            context[key] = now



"""ScriptEngine task module.

Provides the base class for all tasks.
"""

import logging
import uuid
import yaml

import scriptengine.jinja
from scriptengine.yaml.noparse_strings import (
    NoParseJinjaString,
    NoParseYamlString,
)
import scriptengine.yaml

from scriptengine.exceptions import ScriptEngineTaskArgumentInvalidError, \
                                    ScriptEngineTaskArgumentMissingError


_SENTINEL = object()


class Task:

    _reg_name = None
    _invalid_arguments = ('run', 'id', )

    @classmethod
    def check_arguments(cls, arguments):
        """Checks arguments against class members '_invalid_arguments' and
        '_required_arguments', if present. Returns None or throws either
        ScriptEngineTaskArgumentInvalidError or
        ScriptEngineTaskArgumentMissingError.
        """
        for name in arguments:
            if name in getattr(cls, '_invalid_arguments', ()):
                logging.getLogger('se.task').error(
                    (
                        f'Invalid argument "{name}" found while '
                        f'trying to create "{cls.__name__}" task'
                    ),
                    extra={'id': 'no id', 'type': cls.__name__},
                )
                raise ScriptEngineTaskArgumentInvalidError

        for name in getattr(cls, '_required_arguments', ()):
            if name not in arguments:
                logging.getLogger('se.task').error(
                    (
                        f'Missing required argument "{name}" while '
                        f'trying to create "{cls.__name__}" task'
                    ),
                    extra={'id': 'no id', 'type': cls.__name__},
                )
                raise ScriptEngineTaskArgumentMissingError

    def __init__(self, arguments=None):

        self._identifier = uuid.uuid4()

        if arguments is not None:
            Task.check_arguments(arguments)
            for name, value in arguments.items():
                if hasattr(self, name):
                    self.log_error(
                        f'Invalid (reserved name) task argument: {name}'
                    )
                    raise ScriptEngineTaskArgumentInvalidError
                setattr(self, name, value)
        self.log_debug(f'Created task: {self}')

    @classmethod
    def register_name(cls, name):
        cls._reg_name = name

    @property
    def reg_name(self):
        return self._reg_name or \
               f'{self.__class__.__module__}:{self.__class__.__name__}'

    @property
    def id(self):
        return self._identifier

    @property
    def shortid(self):
        return self._identifier.hex[:10]

    def __repr__(self):
        params = {key: val for key, val in self.__dict__.items()
                  if not (isinstance(key, str) and key.startswith('_'))}
        params_list = f'{", ".join([f"{k}={params[k]}" for k in params])}'
        return f'{self.__class__.__name__}({params_list})'

    def run(self, context):
        raise NotImplementedError(
            'Base class function Task.run() must not be called'
        )

    def getarg(self, name, context={}, *,
               parse_jinja=True, parse_yaml=True, default=_SENTINEL):
        """Returns the value of argument 'name'.
           The argument value is parsed with Jinja2 and the given context,
           unless 'parse_jinja' is False. The result is parsed once more with
           the YAML parser in order to get a correctly typed result. If
           parse_yaml is not True, the extra YAML parsing is skipped.
           Parsing with Jinja/YAML is also skipped, if the argument value is a
           string that starts with '_noparse_', '_noparsejinja_',
           '_noparseyaml_', respectively.
           If argument 'name' does not exist, the function raises an
           AttributeError, unless a 'default' value is given.
           """

        def parse(arg_):

            # Recursively parse list items
            if isinstance(arg_, list):
                return [parse(item) for item in arg_]

            # Recursively parse dict values (not keys!)
            if isinstance(arg_, dict):
                return dict((key, parse(val)) for key, val in arg_.items())

            if isinstance(arg_, str):
                if parse_jinja and \
                   not isinstance(arg_, NoParseJinjaString):
                    # Make sure that a NoParseString is still a NoParseString
                    # after this!
                    arg_ = type(arg_)(scriptengine.jinja.render(arg_, context))

                if parse_yaml and \
                   not isinstance(arg_, NoParseYamlString):
                    try:
                        return yaml.full_load(arg_)
                    except (
                        yaml.scanner.ScannerError,
                        yaml.parser.ParserError,
                        yaml.constructor.ConstructorError,
                    ):
                        self.log_debug(
                            f'Reparsing argument "{arg_}" with YAML failed'
                        )

                # Return plain strings, not NoParse*Strings
                return str(arg_)

            # If not a list, dict or string, just return
            return arg_

        try:
            arg = getattr(self, name)
        except AttributeError:
            if default is _SENTINEL:
                self.log_error(
                    f'Trying to access missing task argument: {name}'
                )
                raise ScriptEngineTaskArgumentMissingError
            arg = default
        return parse(arg)

    def _log(self, level, msg):
        logger = logging.getLogger('se.task')
        logger.log(
            level,
            msg,
            extra={'type': self.reg_name, 'id': self.shortid}
        )

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""Context task for ScriptEngine."""

import yaml

from scriptengine.context import ContextUpdate
from scriptengine.exceptions import ScriptEngineTaskError, ScriptEngineTaskRunError
from scriptengine.tasks.core import Task, timed_runner


class Context(Task):
    """Context task, sets/updates the SE context.
    The context task takes name, value pairs as arguments and updates the SE
    context accordingly. Values can be scalars, lists, dictionaries and nested
    combinations thereof.
    Note that Context.run() returns a context *update* and that the higher-level
    running instance (a Job or a ScriptEngine instance) is responsible for
    merging the update with an existing context dict."""

    @timed_runner
    def run(self, context):
        context_update = ContextUpdate(
            {n: self.getarg(n, context) for n in vars(self) if not n.startswith("_")}
        )
        self.log_info(f"Context update: {context_update}")
        return context_update


class ContextFrom(Task):
    """
    This task updates the context, just as the Context task, but from sources
    other than direct arguments.

    If the 'dict' argument is given, the argument value must be a dict and is
    used to update the context, e.g.

    - base.context:
        upd:
            foo: 1
    - base.context.from:
        dict: "{{upd}}"

    If the 'file' argument is given, the argument value must be a file name and
    the context is updated from the file context. The file format must be YAML:

    - base.context.from:
        file: update.yml
    """

    @timed_runner
    def run(self, context):

        dict_arg = self.getarg("dict", context, default=False)
        file_arg = self.getarg("file", context, default=False)

        if dict_arg and file_arg:
            self.log_error("Task arguments 'dict' and 'file' are mutual exclusive")
            raise ScriptEngineTaskError

        if dict_arg:
            self.log_info(f"Context update from dict: {dict_arg}")
            if not isinstance(dict_arg, dict):
                self.log_error(
                    "The 'dict' argument must be a dictionary "
                    f"(was  a '{type(dict_arg).__name__}')"
                )
                raise ScriptEngineTaskRunError
            context_update_d = dict_arg

        elif file_arg:
            self.log_info(f"Context update from file: {file_arg}")
            try:
                with open(file_arg) as f:
                    dict_from_file = yaml.load(f, Loader=yaml.SafeLoader)
            except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
                self.log_error(e)
                raise ScriptEngineTaskRunError
            except yaml.scanner.ScannerError as e:
                self.log_error("Error reading the file as YAML, error message follows:")
                self.log_error(f"  {''.join(str(e).splitlines())}")
                self.log_error(
                    "Note that only YAML file are supported by base.context.from"
                )
                raise ScriptEngineTaskRunError
            if not isinstance(dict_from_file, dict):
                self.log_error(
                    "Reading YAML file succeeded, but the content must be a dict "
                    f"(was a '{type(dict_from_file).__name__}')"
                )
                raise ScriptEngineTaskRunError
            context_update_d = dict_from_file

        else:
            self.log_error("Missing argument: either 'dict' or 'file' is needed")
            raise ScriptEngineTaskError

        return ContextUpdate(context_update_d)



import codecs
import os

import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]


setuptools.setup(
    name="scriptengine",
    version=get_version("scriptengine/version.py"),
    author="Uwe Fladrich",
    author_email="uwe.fladrich@protonmail.com",
    description="A lightweight and extensible framework for executing "
    "scripts written in YAML",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/uwefladrich/scriptengine",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "python-dateutil",
        "deepmerge",
        "deepdiff>=5.7.0,<6.2.0",
        "PyYAML",
        "jinja2",
    ],
    entry_points={
        "console_scripts": [
            "se = scriptengine.cli.se:main",
        ],
        "scriptengine.tasks": [
            # Short aliases for backward compatibility
            "chdir = scriptengine.tasks.base.chdir:Chdir",
            "command = scriptengine.tasks.base.command:Command",
            "context = scriptengine.tasks.base.context:Context",
            "copy = scriptengine.tasks.base.file.copy:Copy",
            "echo = scriptengine.tasks.base.echo:Echo",
            "exit = scriptengine.tasks.base.exit:Exit",
            "find = scriptengine.tasks.base.find:Find",
            "include = scriptengine.tasks.base.include:Include",
            "link = scriptengine.tasks.base.file.link:Link",
            "make_dir = scriptengine.tasks.base.file.make_dir:MakeDir",
            "move = scriptengine.tasks.base.file.move:Move",
            "remove = scriptengine.tasks.base.file.remove:Remove",
            "task_timer = scriptengine.tasks.base.task_timer:TaskTimer",
            "template = scriptengine.tasks.base.template:Template",
            "time = scriptengine.tasks.base.time:Time",
            # Real names
            "base.chdir = scriptengine.tasks.base.chdir:Chdir",
            "base.command = scriptengine.tasks.base.command:Command",
            "base.context = scriptengine.tasks.base.context:Context",
            "base.context.from = scriptengine.tasks.base.context:ContextFrom",
            "base.copy = scriptengine.tasks.base.file.copy:Copy",
            "base.echo = scriptengine.tasks.base.echo:Echo",
            "base.exit = scriptengine.tasks.base.exit:Exit",
            "base.find = scriptengine.tasks.base.find:Find",
            "base.getenv = scriptengine.tasks.base.envvars:Getenv",
            "base.include = scriptengine.tasks.base.include:Include",
            "base.link = scriptengine.tasks.base.file.link:Link",
            "base.make_dir = scriptengine.tasks.base.file.make_dir:MakeDir",
            "base.move = scriptengine.tasks.base.file.move:Move",
            "base.remove = scriptengine.tasks.base.file.remove:Remove",
            "base.setenv = scriptengine.tasks.base.envvars:Setenv",
            "base.task_timer = scriptengine.tasks.base.task_timer:TaskTimer",
            "base.template = scriptengine.tasks.base.template:Template",
            "base.time = scriptengine.tasks.base.time:Time",
        ],
    },
)



""" ScriptEngine logging.
"""

import logging
import logging.config

from scriptengine.helpers import terminal_colors as tc


class LogLevelFormatter:

    def __init__(self, formatters):
        self._formatters = formatters

    def format(self, record):
        return self._formatters[record.levelno].format(record)


def configure(log_level=logging.INFO):

    levels = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    colors = {
        logging.DEBUG: tc.CYAN,
        logging.INFO: tc.GREEN,
        logging.WARNING: tc.ORANGE,
        logging.ERROR: tc.RED,
        logging.CRITICAL: tc.RED,
    }

    formats = {
        'basic': {
            lvl: f'{colors[lvl]}{{asctime}} '
                 f'{tc.BOLD}{{levelname}}{tc.RESET}{colors[lvl]} '
                 f'[{{name}}] {{message}}'
                 f'{tc.RESET}'
            for lvl in levels
        },
        'with_id': {
            lvl: f'{colors[lvl]}{{asctime}} '
                 f'{tc.BOLD}{{levelname}}{tc.RESET}{colors[lvl]} '
                 f'[{{name}} <{{id}}>] {{message}}'
                 f'{tc.RESET}'
            for lvl in levels
        },
        'with_id_and_type': {
            lvl: f'{colors[lvl]}{{asctime}} '
                 f'{tc.BOLD}{{levelname}}{tc.RESET}{colors[lvl]} '
                 f'[{{name}}:{{type}} <{{id}}>] {{message}}'
                 f'{tc.RESET}'
            for lvl in levels
        },
    }

    date_format = '%Y-%m-%d %H:%M:%S'

    configuration = dict(

        version=1,

        formatters={
            'basic': {
                '()': 'scriptengine.logging.LogLevelFormatter',
                'formatters': {
                    lvl: logging.Formatter(
                            fmt=formats['basic'][lvl],
                            style='{',
                            datefmt=date_format
                         )
                    for lvl in levels
                }
            },
            'job': {
                '()': 'scriptengine.logging.LogLevelFormatter',
                'formatters': {
                    lvl: logging.Formatter(
                            fmt=formats['with_id'][lvl],
                            style='{',
                            datefmt=date_format
                         )
                    for lvl in levels
                }
            },
            'task': {
                '()': 'scriptengine.logging.LogLevelFormatter',
                'formatters': {
                    lvl: logging.Formatter(
                            fmt=formats['with_id_and_type'][lvl],
                            style='{',
                            datefmt=date_format
                         )
                    for lvl in levels
                }
            },
        },

        handlers={
            'default': {
                'class': 'logging.StreamHandler',
                'formatter': 'basic',
                },
            'job': {
                'class': 'logging.StreamHandler',
                'formatter': 'job',
                },
            'task': {
                'class': 'logging.StreamHandler',
                'formatter': 'task',
                },
        },

        loggers={
            'se': {
                'handlers': ['default'],
                'level': log_level,
            },
            'se.job': {
                'handlers': ['job'],
                'level': log_level,
                'propagate': False,
            },
            'se.task': {
                'handlers': ['task'],
                'level': log_level,
                'propagate': False,
            },
            'se.task.loader': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': False,
            },
        },
    )
    logging.config.dictConfig(configuration)



""" ScriptEngine jobs.

Jobs are the second level (above tasks) work unit in ScriptEngine. Jobs are
made of tasks and may include loops, conditionals, and a context. Job lists can
be given to a ScriptEngine, for execution.
"""

import ast
import logging
import uuid

from scriptengine.context import Context, ContextUpdate, save_copy
from scriptengine.exceptions import (
    ScriptEngineJobParseError,
    ScriptEngineParseJinjaError,
)
from scriptengine.jinja import render as j2render
from scriptengine.tasks.core import Task


def _listy(thing, type_=list):
    if isinstance(thing, str) or isinstance(thing, dict):
        return type_((thing,))
    try:
        iter(thing)
    except TypeError:
        return type_((thing,))
    else:
        return type_(thing)


def _is_todo(thing):
    return isinstance(thing, Task) or isinstance(thing, Job)


def _todo_list(todo):
    todo_list = _listy(todo)
    if not all(_is_todo(t) for t in todo_list):
        raise ValueError("A todo list must be made of Tasks and/or Jobs")
    return todo_list


class Job:
    def __init__(self, todo=None, *, when=None, loop=None, loop_vars=None):
        self._identifier = uuid.uuid4()
        self.todo = todo or []
        self._when = when
        self._loop = loop
        self._loop_vars = loop_vars
        self.log_debug(
            "New Job:"
            f"{' when '+self._when if self._when else ''}"
            f"{' with '+str(self._loop_vars) if self._loop else ''}"
            f"{' in '+str(self._loop) if self._loop else ''}"
            f" {tuple(t.shortid for t in self._todo)}"
        )

    @property
    def id(self):
        return self._identifier

    @property
    def shortid(self):
        return self._identifier.hex[:10]

    @property
    def todo(self):
        return self._todo

    @todo.setter
    def todo(self, todo):
        self._todo = _todo_list(todo)

    def when(self, context):
        try:
            return self._when is None or j2render(self._when, context, boolean=True)
        except ScriptEngineParseJinjaError:
            self.log_error(
                "Error while parsing (Jinja2) invalid when clause "
                f'"{self._when}" with context "{context}"'
            )
            raise ScriptEngineJobParseError

    def loop_spec(self, context):
        try:
            iter = j2render(self._loop, context)
        except ScriptEngineParseJinjaError:
            self.log_error(
                "Error while parsing (Jinja2) invalid loop expression "
                f'"{self._loop}" with context "{context}"'
            )
            raise ScriptEngineJobParseError
        if isinstance(iter, str):
            try:
                iter = ast.literal_eval(iter or "None")
            except (SyntaxError, ValueError):
                self.log_error(
                    "Error while evaluating (AST) invalid loop expression "
                    f'"{iter}" with context "{context}"'
                )
                raise ScriptEngineJobParseError
        if isinstance(iter, dict):
            iter = iter.items()
            vars = self._loop_vars or ("key", "value")
        else:
            vars = self._loop_vars or ("item",)
        return iter, vars

    def loop(self, context):
        if self._loop:
            iter, vars = self.loop_spec(context)
            if iter:
                for items in iter:
                    parsed_items = tuple(
                        j2render(i, context) for i in _listy(items, type_=tuple)
                    )
                    vars_tuple = _listy(vars, type_=tuple)
                    if len(vars_tuple) == 1 and len(parsed_items) > 1:
                        # If only one loop variable is given and items is listy,
                        # assign the whole list to the variable (see #59)
                        yield {vars_tuple[0]: parsed_items}
                    else:
                        # Otherwise, map (zip) loop vars with items
                        # Note that extra vars or items are ignored!
                        yield dict(zip(vars_tuple, parsed_items))
            else:
                self.log_warning(
                    "Null loop after parsing loop descriptor, job is not run!"
                )
                raise StopIteration
        else:
            yield {}

    def append(self, todo):
        todo_list = _todo_list(todo)
        self._todo.extend(todo_list)
        self.log_debug(f'Append: {",".join(t.shortid for t in todo_list)}')

    def run(self, context):
        if self.when(context):
            local_context = Context(save_copy(context))
            for items in self.loop(local_context):
                if set(items) & set(local_context):
                    self.log_warning(
                        "The following loop variables collide with the "
                        f"context: {set(items) & set(local_context)}"
                    )
                for t in self.todo:
                    context_update = t.run(Context({**local_context, **items}))
                    if context_update:
                        local_context += context_update
            return ContextUpdate(context, local_context)

    def _log(self, level, msg):
        logging.getLogger("se.job").log(level, msg, extra={"id": self.shortid})

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""Template task for ScriptEngine."""

from pathlib import Path
import os
import stat
import jinja2

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.jinja import render as j2render
from scriptengine.jinja import filters as j2filters


class Template(Task):
    """Template task, renders a template with Jinja2 and writes result to file.

    The task needs a source (a Jinja2 template file) and a destination. It
    renders the template given the context passed in kwargs and writes the
    result to the destination file.

    Args:
        dictionary (dict): Must at least contain the following keys:
            - src: Source file name (a Jinja2 template)
            - dst: Destination file name
    """

    _required_arguments = (
        "src",
        "dst",
    )

    def __init__(self, arguments):
        Template.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f"Template: {self.src} --> {self.dst}"

    @timed_runner
    def run(self, context):
        src = self.getarg("src", context)
        dst = Path(self.getarg("dst", context))
        self.log_info(f"Jinja2 render: {src} --> {dst}")

        # Prepare the Jinja render environment with the proper search path
        search_path = (
            Path("."),
            Path(".") / "templates",
            Path(context["se"]["cli"]["cwd"]),
            Path(context["se"]["cli"]["cwd"]) / "templates",
        )
        self.log_debug(f"Template search path: {tuple(map(str, search_path))}")
        j2loader = jinja2.FileSystemLoader(search_path)
        j2env = jinja2.Environment(loader=j2loader)
        for name, function in j2filters().items():
            j2env.filters[name] = function

        output = j2render(j2env.get_template(src).render(context), context)
        with dst.open(mode="w") as f:
            f.write(output + "\n")

        if self.getarg("executable", context, default=False):
            umask = os.umask(0)
            os.umask(umask)
            dst.chmod(
                dst.stat().st_mode
                | (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) & ~umask
            )
